Ns = 2.546899e19  # [cm^-3] molecular number density for standard air conditions
ps = 1013.25  # [hPa]   standard air pressure
Ts = 288.15  # [K]     standard air temperature
Hs = 9500.0  # [km]    density scale hight for La Palma Winter
