# standard atmosphere thermodynamic values
STD_NUMBER_DENSITY = 2.546899e19  # [cm^-3] molecular number density for standard air conditions
STD_AIR_PRESSURE = 1013.25  # [hPa]   standard air pressure
STD_AIR_TEMPERATURE = 288.15  # [K]     standard air temperature
DENSITY_SCALE_HEIGHT = 9500.0  # [km]    density scale hight for La Palma Winter
STD_RELATIVE_HUMIDITY = 45.9  # [%]     standard air rel. humidity
# atmospheric composition
NITROGEN_RATIO = 0.78084
OXYGEN_RATIO = 0.20946
ARGON_RATIO = 0.00934
GAS_CONSTANT = 8.31451  # gas constant [J/mol/K]
MOLAR_MASS_WATER_VAPOR = 0.018015  # molar mass of water vapor [kg/mol]
